For the following object please execute the following transformation. The output you provide should be the same type as the object provided:
Object:
{{ content }}

Transformation:
{{ transformation }}