For the following table please execute the following transformation. The table shown here is a representation of the actual `Table` object any tools will be run against.

The output you provide should be a markdown table or a function call that results in a `Table` object. 

If using a tool, use the tool whose name most closely matches the transformation request:
Table:
{{ table }}

Transformation:
{{ transformation }}
