Your task is to imitate the output of the following function for the given arguments.
Reply Nothing else but the output of the function.

Function:
def {{ function.name }}({{ function.signature }})
    """{{ function.docstring | default("No documentation provided.") }}"""

{%- if arguments|length > 0 %}

Arguments:
{%- for arg in arguments %}
- {{ arg.name }}: {{ arg.value }}  (type: {{ arg.annotation -}})
{%- endfor -%}
{%- endif -%}