{% block prefix %}
{%- if prefix -%}
  {{- prefix }}
{% endif -%}
{% endblock prefix %}

{%- block description %}
{%- if description -%}
  {{- description -}}
{%- endif -%}
{% endblock description %}

{%- block requirements -%}
{%- if requirements|length > 0 %}
Fulfill the following requirements:
{%- for req in requirements %}
* {{ req -}}
{% endfor %}
{%- endif -%}
{% endblock requirements %}

{%- block icl_examples -%}
{%- if icl_examples|length > 0 %}
Here are some examples:
{%- for example in icl_examples %}
* {{ example -}}
{% endfor %}
{%- endif -%}
{% endblock icl_examples %}

{%- block grounding_context -%}
{%- if grounding_context|length > 0 %}
Here is some grounding context:
{%- for key, value in grounding_context.items() %}
*  {% if key %} [{{ key }}] = {% endif %} {{ value -}}
{% endfor %}
{%- endif -%}
{% endblock grounding_context %}

{%- block repair_block -%}
{% if repair %}
{{ repair -}}
{%- endif -%}
{% endblock repair_block %}

{%- block output_prefix -%}
{% if output_prefix %}
{{ output_prefix -}}
{%- endif -%}
{% endblock output_prefix %}
