Please check the following requirement against the following output.
Reply with 'yes' if the requirement is satisfied and 'no' otherwise.
Do not include any other text in your response.

{%- block output -%}
{% if output %}
  Output: {{ output -}}
{%- endif -%}
{% endblock output%}

{%- block description -%}
{% if description %}
  Requirement: {{ description -}}
{%- endif -%}
{% endblock description %}
